import datetime as dt
import json
import warnings

import numpy as np
from pandas import DataFrame, DatetimeIndex, MultiIndex, Series, concat, to_datetime
from pandas.io.json import read_json
from pandas.tseries.offsets import MonthEnd

from pandas_datareader._utils import RemoteDataError
from pandas_datareader.base import _OptionBaseReader

# Items needed for options class
CUR_MONTH = dt.datetime.now().month
CUR_YEAR = dt.datetime.now().year
CUR_DAY = dt.datetime.now().day


def _parse_options_data(jd):
    return read_json(jd)


class Options(_OptionBaseReader):
    """
    ***Experimental***
    This class fetches call/put data for a given stock/expiry month.

    It is instantiated with a string representing the ticker symbol.

    The class has the following methods:
        get_options_data(month, year, expiry)
        get_call_data(month, year, expiry)
        get_put_data(month, year, expiry)
        get_near_stock_price(opt_frame, above_below)
        get_all_data(call, put)
        get_forward_data(months, call, put) (deprecated)

    Examples
    --------
    # Instantiate object with ticker
    >>> aapl = Options('aapl')

    # Fetch next expiry call data
    >>> calls = aapl.get_call_data()

    # Can now access aapl.calls instance variable
    >>> aapl.calls

    # Fetch next expiry put data
    >>> puts = aapl.get_put_data()

    # Can now access aapl.puts instance variable
    >>> aapl.puts

    # cut down the call data to be 3 below and 3 above the stock price.
    >>> cut_calls = aapl.get_near_stock_price(call=True, above_below=3)

    # Fetch call and put data with expiry from now to 8 months out
    >>> forward_data = aapl.get_forward_data(8, call=True, put=True)

    # Fetch all call and put data
    >>> all_data = aapl.get_all_data()
    """

    _OPTIONS_BASE_URL = "https://query1.finance.yahoo.com/" "v7/finance/options/{sym}"

    def get_options_data(self, month=None, year=None, expiry=None):
        """
        ***Experimental***
        Gets call/put data for the stock with the expiration data in the
        given month and year

        Parameters
        ----------
        month : number, int, optional(default=None)
            The month the options expire. This should be either 1 or 2
            digits.

        year : number, int, optional(default=None)
            The year the options expire. This should be a 4 digit int.

        expiry : date-like or convertible or
                 list-like object, optional (default=None)
            The date (or dates) when options expire (defaults to current month)

        Returns
        -------
        pandas.DataFrame
            A DataFrame with requested options data.

            Index:
                Strike: Option strike, int
                Expiry: Option expiry, Timestamp
                Type: Call or Put, string
                Symbol: Option symbol as reported on Yahoo, string
            Columns:
                Last: Last option price, float
                Chg: Change from prior day, float
                PctChg: Change from prior day in percent, float
                Bid: Bid price, float
                Ask: Ask price, float
                Vol: Volume traded, int64
                Open_Int: Open interest, int64
                IsNonstandard: True if the the deliverable is not 100 shares,
                               otherwise False
                Underlying: Ticker of the underlying security, string
                Underlying_Price: Price of the underlying security, float64
                Quote_Time: Time of the quote, Timestamp
                Last_Trade_Date: Time of the last trade for this expiry
                                 and strike, Timestamp
                IV: Implied volatility, float
                JSON: Parsed json object, json
                    Useful to extract other returned key/value pairs as needed

        Notes
        -----
        Note: Format of returned DataFrame is dependent
              on Yahoo and may change.

        When called, this function will add instance variables named
        calls and puts. See the following example:

            >>> aapl = Options('aapl', 'yahoo')  # Create object
            >>> aapl.calls  # will give an AttributeError
            >>> aapl.get_options()  # Get data and set ivars
            >>> aapl.calls  # Doesn't throw AttributeError

        Also note that aapl.calls and appl.puts will always be the calls
        and puts for the next expiry. If the user calls this method with
        a different expiry, the ivar will be named callsYYMMDD or putsYYMMDD,
        where YY, MM and DD are, respectively, two digit representations of
        the year, month and day for the expiry of the options.

        """
        return concat(
            [f(month, year, expiry) for f in (self.get_put_data, self.get_call_data)]
        ).sort_index()

    def _option_from_url(self, url):

        jd = self._parse_url(url)
        result = jd["optionChain"]["result"]
        try:
            calls = result["options"]["calls"]
            puts = result["options"]["puts"]
        except IndexError:
            raise RemoteDataError("Option json not available " "for url: %s" % url)

        self.underlying_price = (
            result["quote"]["regularMarketPrice"]
            if result["quote"]["marketState"] == "PRE"
            else result["quote"]["preMarketPrice"]
        )
        quote_unix_time = (
            result["quote"]["regularMarketTime"]
            if result["quote"]["marketState"] == "PRE"
            else result["quote"]["preMarketTime"]
        )
        self.quote_time = dt.datetime.fromtimestamp(quote_unix_time)

        calls = _parse_options_data(calls)
        puts = _parse_options_data(puts)

        calls = self._process_data(calls)
        puts = self._process_data(puts)

        return {"calls": calls, "puts": puts}

    def _get_option_data(self, expiry, name):
        frame_name = "_frames" + self._expiry_to_string(expiry)

        try:
            frames = getattr(self, frame_name)
        except AttributeError:
            frames = self._get_option_frames_from_yahoo(expiry)

        option_data = frames[name]
        if expiry != self.expiry_dates[0]:
            name += self._expiry_to_string(expiry)

        setattr(self, name, option_data)
        return option_data

    def get_call_data(self, month=None, year=None, expiry=None):
        """
        ***Experimental***
        Gets call/put data for the stock with the expiration data in the
        given month and year

        Parameters
        ----------
        month : number, int, optional(default=None)
            The month the options expire. This should be either 1 or 2
            digits.

        year : number, int, optional(default=None)
            The year the options expire. This should be a 4 digit int.

        expiry : date-like or convertible or
                 list-like object, optional (default=None)
            The date (or dates) when options expire (defaults to current month)

        Returns
        -------
        call_data: pandas.DataFrame
            A DataFrame with requested options data.

            Index:
                Strike: Option strike, int
                Expiry: Option expiry, Timestamp
                Type: Call or Put, string
                Symbol: Option symbol as reported on Yahoo, string
            Columns:
                Last: Last option price, float
                Chg: Change from prior day, float
                PctChg: Change from prior day in percent, float
                Bid: Bid price, float
                Ask: Ask price, float
                Vol: Volume traded, int64
                Open_Int: Open interest, int64
                IsNonstandard: True if the the deliverable is not 100 shares,
                               otherwise false
                Underlying: Ticker of the underlying security, string
                Underlying_Price: Price of the underlying security, float64
                Quote_Time: Time of the quote, Timestamp
                Last_Trade_Date: Time of the last trade for this expiry
                                 and strike, Timestamp
                IV: Implied volatility, float
                JSON: Parsed json object, json
                    Useful to extract other returned key/value pairs as needed

        Notes
        -----
        Note: Format of returned DataFrame is dependent
              on Yahoo and may change.

        When called, this function will add instance variables named
        calls and puts. See the following example:

            >>> aapl = Options('aapl', 'yahoo')  # Create object
            >>> aapl.calls  # will give an AttributeError
            >>> aapl.get_call_data()  # Get data and set ivars
            >>> aapl.calls  # Doesn't throw AttributeError

        Also note that aapl.calls will always be the calls for the next
        expiry. If the user calls this method with a different month
        or year, the ivar will be named callsYYMMDD where YY, MM and DD are,
        respectively, two digit representations of the year, month and day
        for the expiry of the options.
        """
        expiry = self._try_parse_dates(year, month, expiry)
        return self._get_data_in_date_range(expiry, call=True, put=False)

    def get_put_data(self, month=None, year=None, expiry=None):
        """
        ***Experimental***
        Gets put data for the stock with the expiration data in the
        given month and year

        Parameters
        ----------
        month : number, int, optional(default=None)
            The month the options expire. This should be either 1 or 2
            digits.

        year : number, int, optional(default=None)
            The year the options expire. This should be a 4 digit int.

        expiry : date-like or convertible or
                 list-like object, optional (default=None)
            The date (or dates) when options expire (defaults to current month)

        Returns
        -------
        put_data: pandas.DataFrame
            A DataFrame with requested options data.

            Index:
                Strike: Option strike, int
                Expiry: Option expiry, Timestamp
                Type: Call or Put, string
                Symbol: Option symbol as reported on Yahoo, string
            Columns:
                Last: Last option price, float
                Chg: Change from prior day, float
                PctChg: Change from prior day in percent, float
                Bid: Bid price, float
                Ask: Ask price, float
                Vol: Volume traded, int64
                Open_Int: Open interest, int64
                IsNonstandard: True if the the deliverable is not 100 shares,
                               otherwise false
                Underlying: Ticker of the underlying security, string
                Underlying_Price: Price of the underlying security, float64
                Quote_Time: Time of the quote, Timestamp
                Last_Trade_Date: Time of the last trade for this expiry
                                 and strike, Timestamp
                IV: Implied volatility, float
                JSON: Parsed json object, json
                    Useful to extract other returned key/value pairs as needed

        Notes
        -----
        Note: Format of returned DataFrame is dependent
              on Yahoo and may change.

        When called, this function will add instance variables named
        puts. See the following example:

            >>> aapl = Options('aapl')  # Create object
            >>> aapl.puts  # will give an AttributeError
            >>> aapl.get_put_data()  # Get data and set ivars
            >>> aapl.puts  # Doesn't throw AttributeError

                    return self.__setattr__(self, str(str(x) + str(y)))

        Also note that aapl.puts will always be the puts for the next
        expiry. If the user calls this method with a different month
        or year, the ivar will be named putsYYMMDD where YY, MM and DD are,
        respectively, two digit representations of the year, month and day
        for the expiry of the options.
        """
        expiry = self._try_parse_dates(year, month, expiry)
        return self._get_data_in_date_range(expiry, put=True, call=False)

    def get_near_stock_price(
        self, above_below=2, call=True, put=False, month=None, year=None, expiry=None
    ):
        """
        ***Experimental***
        Returns a DataFrame of options that are near the current stock price.

        Parameters
        ----------
        above_below : number, int, optional (default=2)
            The number of strike prices above and below the stock price that
            should be taken

        call : bool
            Tells the function whether or not it should be using calls

        put : bool
            Tells the function weather or not it should be using puts

        month : number, int, optional(default=None)
            The month the options expire. This should be either 1 or 2
            digits.

        year : number, int, optional(default=None)
            The year the options expire. This should be a 4 digit int.

        expiry : date-like or convertible or list-like object,
                 optional (default=None)
            The date (or dates) when options expire (defaults to current month)

        Returns
        -------
        chopped: DataFrame
            The resultant DataFrame chopped down to be 2 * above_below + 1 rows
            desired. If there isn't data as far out as the user has asked for
            then

            Index:
                Strike: Option strike, int
                Expiry: Option expiry, Timestamp
                Type: Call or Put, string
                Symbol: Option symbol as reported on Yahoo, string
            Columns:
                Last: Last option price, float
                Chg: Change from prior day, float
                PctChg: Change from prior day in percent, float
                Bid: Bid price, float
                Ask: Ask price, float
                Vol: Volume traded, int64
                Open_Int: Open interest, int64
                IsNonstandard: True if the the deliverable is not 100 shares,
                               otherwise false
                Underlying: Ticker of the underlying security, string
                Underlying_Price: Price of the underlying security, float64
                Quote_Time: Time of the quote, Timestamp
                Last_Trade_Date: Time of the last trade for this expiry
                                 and strike, Timestamp
                IV: Implied volatility, float
                JSON: Parsed json object, json
                    Useful to extract other returned key/value pairs as needed

         Note: Format of returned DataFrame is dependent
               on Yahoo and may change.

        """
        expiry = self._try_parse_dates(year, month, expiry)
        data = self._get_data_in_date_range(expiry, call=call, put=put)
        underlying_price = data.Underlying_Price[0]
        return self._chop_data(data, above_below, underlying_price)

    def _chop_data(self, df, above_below=2, underlying_price=None):
        """
        Returns a DataFrame only options that are near the current stock price.
        """

        if not underlying_price:
            try:
                underlying_price = self.underlying_price
            except AttributeError:
                underlying_price = np.nan

        max_strike = max(df.index.get_level_values("Strike"))
        min_strike = min(df.index.get_level_values("Strike"))

        if (
            not np.isnan(underlying_price)
            and min_strike < underlying_price < max_strike
        ):
            start_index = np.where(
                df.index.get_level_values("Strike") > underlying_price
            )[0][0]

            get_range = slice(start_index - above_below, start_index + above_below + 1)
            df = df[get_range].dropna(how="all")

        return df

    def _try_parse_dates(self, year, month, expiry):
        """
        Validates dates provided by user. Ensures the user either provided
        both a month and a year or an expiry.

        Parameters
        ----------
        year : int
            Calendar year

        month : int
            Calendar month

        expiry : date-like or convertible, (preferred)
            Expiry date

        Returns
        -------
        list of expiry dates (datetime.date)
        """

        # Checks if the user gave one of the month or the year
        # but not both and did not provide an expiry:
        if (
            (month is not None and year is None)
            or (month is None and year is not None)
            and expiry is None
        ):
            msg = (
                "You must specify either (`year` and `month`) or `expiry` "
                "or none of these options for the next expiry."
            )
            raise ValueError(msg)

        if expiry is not None:
            if hasattr(expiry, "__iter__"):
                expiry = [self._validate_expiry(exp) for exp in expiry]
            else:
                expiry = [self._validate_expiry(expiry)]

            if len(expiry) == 0:
                raise ValueError("No expiries available for given input.")

        elif year is None and month is None:
            # No arguments passed, provide next expiry
            year = CUR_YEAR
            month = CUR_MONTH
            expiry = dt.date(year, month, 1)
            expiry = [self._validate_expiry(expiry)]

        else:
            # Year and month passed, provide all expiries in that month
            expiry = [
                expiry
                for expiry in self.expiry_dates
                if expiry.year == year and expiry.month == month
            ]
            if len(expiry) == 0:
                raise ValueError("No expiries available " "in %s-%s" % (year, month))

        return expiry

    def _validate_expiry(self, expiry):
        """
        Ensures that an expiry date has data available on Yahoo.

        If the expiry date does not have options that expire on that day,
        return next expiry.
        """

        expiry_dates = self.expiry_dates
        expiry = to_datetime(expiry)
        if hasattr(expiry, "date"):
            expiry = expiry.date()

        if expiry in expiry_dates:
            return expiry
        else:
            index = DatetimeIndex(expiry_dates).sort_values()
            return index[index.date >= expiry][0].date()

    def get_forward_data(
        self, months, call=True, put=False, near=False, above_below=2
    ):  # pragma: no cover
        """
        ***Experimental***
        Gets either call, put, or both data for months starting in the current
        month and going out in the future a specified amount of time.

        Parameters
        ----------
        months : number, int
            How many months to go out in the collection of the data. This is
            inclusive.

        call : bool, optional (default=True)
            Whether or not to collect data for call options

        put : bool, optional (default=False)
            Whether or not to collect data for put options.

        near : bool, optional (default=False)
            Whether this function should get only the data near the
            current stock price. Uses Options.get_near_stock_price

        above_below : number, int, optional (default=2)
            The number of strike prices above and below the stock price that
            should be taken if the near option is set to True

        Returns
        -------
        pandas.DataFrame
            A DataFrame with requested options data.

            Index:
                Strike: Option strike, int
                Expiry: Option expiry, Timestamp
                Type: Call or Put, string
                Symbol: Option symbol as reported on Yahoo, string
            Columns:
                Last: Last option price, float
                Chg: Change from prior day, float
                PctChg: Change from prior day in percent, float
                Bid: Bid price, float
                Ask: Ask price, float
                Vol: Volume traded, int64
                Open_Int: Open interest, int64
                IsNonstandard: True if the the deliverable is not 100 shares,
                               otherwise false
                Underlying: Ticker of the underlying security, string
                Underlying_Price: Price of the underlying security, float64
                Quote_Time: Time of the quote, Timestamp
                Last_Trade_Date: Time of the last trade for this expiry
                                 and strike, Timestamp
                IV: Implied volatility, float
                JSON: Parsed json object, json
                    Useful to extract other returned key/value pairs as needed

                Note: Format of returned DataFrame is dependent
                      on Yahoo and may change.

        """
        warnings.warn("get_forward_data() is deprecated", FutureWarning)
        end_date = dt.date.today() + MonthEnd(months)
        dates = [date for date in self.expiry_dates if date <= end_date.date()]
        data = self._get_data_in_date_range(dates, call=call, put=put)
        if near:
            data = self._chop_data(data, above_below=above_below)
        return data

    def get_all_data(self, call=True, put=True):
        """
        ***Experimental***
        Gets either call, put, or both data for all available months starting
        in the current month.

        Parameters
        ----------
        call : bool, optional (default=True)
            Whether or not to collect data for call options

        put : bool, optional (default=True)
            Whether or not to collect data for put options.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with requested options data.

            Index:
                Strike: Option strike, int
                Expiry: Option expiry, Timestamp
                Type: Call or Put, string
                Symbol: Option symbol as reported on Yahoo, string
            Columns:
                Last: Last option price, float
                Chg: Change from prior day, float
                PctChg: Change from prior day in percent, float
                Bid: Bid price, float
                Ask: Ask price, float
                Vol: Volume traded, int64
                Open_Int: Open interest, int64
                IsNonstandard: True if the the deliverable is not 100 shares,
                               otherwise false
                Underlying: Ticker of the underlying security, string
                Underlying_Price: Price of the underlying security, float64
                Quote_Time: Time of the quote, Timestamp
                Last_Trade_Date: Time of the last trade for this expiry
                                 and strike, Timestamp
                IV: Implied volatility, float
                JSON: Parsed json object, json
                    Useful to extract other returned key/value pairs as needed

        Note: Format of returned DataFrame is dependent
              on Yahoo and may change.
        """

        return self._load_data()

    def _get_data_in_date_range(self, dates, call=True, put=True):

        to_ret = Series({"call": call, "put": put})
        to_ret = to_ret[to_ret].index

        df = self._load_data(dates)
        types = [typ for typ in to_ret]

        df_filtered_by_type = df[df.index.map(lambda x: x[2] in types).tolist()]
        df_filtered_by_expiry = df_filtered_by_type[
            df_filtered_by_type.index.get_level_values("Expiry").isin(dates)
        ]
        return df_filtered_by_expiry

    @property
    def underlying_price(self):
        """
        Returns the underlying price.
        """
        try:
            underlying_price = self._underlying_price
        except AttributeError:
            data = self.get_options_data()
            underlying_price = data.Underlying_Price[0]
        return underlying_price

    @property
    def quote_time(self):
        """
        Returns the quote time.
        """
        try:
            quote_time = self._quote_time
        except AttributeError:
            data = self.get_options_data()
            quote_time = data.Quote_Time[0]
        return quote_time

    @property
    def expiry_dates(self):
        """
        Returns a list of available expiry dates
        """
        try:
            expiry_dates = self._expiry_dates
        except AttributeError:
            expiry_dates = self._get_expiry_dates()
        return expiry_dates

    def _get_expiry_dates(self):
        """
        Gets available expiry dates.

        Returns
        -------
        List of datetime.date objects
        """

        url = self._OPTIONS_BASE_URL.format(sym=self.symbol)
        jd = self._parse_url(url)

        expiry_dates = [
            dt.datetime.utcfromtimestamp(ts).date()
            for ts in jd["optionChain"]["result"][0]["expirationDates"]
        ]

        if len(expiry_dates) == 0:
            raise RemoteDataError("Data not available")  # pragma: no cover

        self._expiry_dates = expiry_dates
        return expiry_dates

    def _parse_url(self, url):
        """

        Downloads and parses a URL into a json object.
        Parameters
        ----------
        url : String
            The url to load and parse

        Returns
        -------
        A JSON object
        """
        jd = json.loads(self._read_url_as_StringIO(url).read())
        if jd is None:  # pragma: no cover
            raise RemoteDataError(
                "Parsed URL {0!r} is not " "a valid json object".format(url)
            )
        return jd

    def _process_data(self, jd):
        """
        Parse JSON data into the DataFrame

        Also adds computed columns for:
            IsNonstandard (ie: deliverable is not 100 shares)

        Parameters
        ----------
        jd : json object
            a json object containing the data to process

        Returns
        -------
        pandas.DataFrame
            A DataFrame with requested options data.
        """

        columns = [
            "Last",
            "Bid",
            "Ask",
            "Chg",
            "PctChg",
            "Vol",
            "Open_Int",
            "IV",
            "Root",
            "IsNonstandard",
            "Underlying",
            "Underlying_Price",
            "Quote_Time",
            "Last_Trade_Date",
            "JSON",
        ]
        indexes = ["Strike", "Expiry", "Type", "Symbol"]
        rows_list, index = self._process_rows(jd)
        if len(rows_list) > 0:
            df = DataFrame(
                rows_list,
                columns=columns,
                index=MultiIndex.from_tuples(index, names=indexes),
            )
        else:
            df = DataFrame(columns=columns)

        df["IsNonstandard"] = df["Root"] != self.symbol.replace("-", "")

        # Make dtype consistent, requires float64 as there can be NaNs
        df["Vol"] = df["Vol"].astype("float64")
        df["Open_Int"] = df["Open_Int"].astype("float64")

        return df.sort_index()

    def _process_rows(self, jd):
        rows_list = []
        index = []

        # handle no results
        if len(jd["optionChain"]["result"]) <= 0:
            return rows_list, index

        quote = jd["optionChain"]["result"][0]["quote"]
        for option in jd["optionChain"]["result"][0]["options"]:
            for typ in ["calls", "puts"]:
                options_by_type = option[typ]
                for option_by_strike in options_by_type:
                    d = {}
                    for dkey, rkey, ntype in [
                        ("Last", "lastPrice", float),
                        ("Bid", "bid", float),
                        ("Ask", "ask", float),
                        ("Chg", "change", float),
                        ("PctChg", "percentChange", float),
                        ("Vol", "volume", int),
                        ("Open_Int", "openInterest", int),
                        ("IV", "impliedVolatility", float),
                        ("Last_Trade_Date", "lastTradeDate", int),
                    ]:
                        try:
                            d[dkey] = ntype(option_by_strike[rkey])
                        except (KeyError, ValueError):
                            pass
                    d["JSON"] = option_by_strike
                    d["Root"] = option_by_strike["contractSymbol"][:-15]
                    d["Underlying"] = self.symbol

                    d["Underlying_Price"] = quote["regularMarketPrice"]
                    quote_unix_time = quote["regularMarketTime"]
                    if quote["marketState"] == "PRE" and "preMarketPrice" in quote:
                        d["Underlying_Price"] = quote["preMarketPrice"]
                        quote_unix_time = quote["preMarketTime"]
                    elif (
                        quote["marketState"] == "POSTPOST"
                        and "postMarketPrice" in quote
                    ):
                        d["Underlying_Price"] = quote["postMarketPrice"]
                        quote_unix_time = quote["postMarketTime"]
                    d["Quote_Time"] = dt.datetime.utcfromtimestamp(quote_unix_time)

                    self._underlying_price = d["Underlying_Price"]
                    self._quote_time = d["Quote_Time"]

                    d["Last_Trade_Date"] = dt.datetime.utcfromtimestamp(
                        d["Last_Trade_Date"]
                    )

                    rows_list.append(d)
                    index.append(
                        (
                            float(option_by_strike["strike"]),
                            dt.datetime.utcfromtimestamp(
                                option_by_strike["expiration"]
                            ),
                            typ.replace("s", ""),
                            option_by_strike["contractSymbol"],
                        )
                    )
        return rows_list, index

    def _load_data(self, exp_dates=None):
        """
        Loads options data.

        Parameters
        ----------
        exp_dates : tuple of datetimes, optional(default=None)
            The expiry dates to load options data for. If none is specified,
            uses all expiry dates available for the symbol.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with requested options data.
        """
        data = []
        epoch = dt.datetime.utcfromtimestamp(0)

        try:
            if exp_dates is None:
                exp_dates = self._get_expiry_dates()
            exp_unix_times = [
                int(
                    (
                        dt.datetime(exp_date.year, exp_date.month, exp_date.day) - epoch
                    ).total_seconds()
                )
                for exp_date in exp_dates
            ]
            for exp_date in exp_unix_times:
                url = (self._OPTIONS_BASE_URL + "?date={exp_date}").format(
                    sym=self.symbol, exp_date=exp_date
                )
                jd = self._parse_url(url)
                data.append(self._process_data(jd))
            return concat(data).sort_index()
        finally:
            self.close()



import pytest
from requests.exceptions import HTTPError

from pandas_datareader import data as web

pytestmark = pytest.mark.stable


class TestMoex(object):
    def test_moex_datareader(self):
        try:
            df = web.DataReader(
                "USD000UTSTOM", "moex", start="2017-07-01", end="2017-07-31"
            )
            assert "SECID" in df.columns
        except HTTPError as e:
            pytest.skip(e)

    def test_moex_stock_datareader(self):
        try:
            df = web.DataReader(
                ["GAZP", "SIBN"], "moex", start="2019-12-26", end="2019-12-26"
            )
            assert len(df) == 2
        except HTTPError as e:
            pytest.skip(e)

    def test_moex_datareader_filter(self):
        try:
            df = web.DataReader("SBER", "moex", start="2020-07-14", end="2020-07-14")
            assert len(df) == 1
        except HTTPError as e:
            pytest.skip(e)



import pytest

from pandas_datareader import data as web
from pandas_datareader.data import get_data_stooq

pytestmark = pytest.mark.stable


def test_stooq_dji():
    f = web.DataReader("GS", "stooq")
    assert f.shape[0] > 0


def test_get_data_stooq_dji():
    f = get_data_stooq("AMZN")
    assert f.shape[0] > 0


def test_get_data_stooq_dates():
    f = get_data_stooq("SPY", start="20180101", end="20180115")
    assert f.shape[0] == 9


def test_stooq_sp500():
    f = get_data_stooq("^SPX")
    assert f.shape[0] > 0


def test_stooq_clx19f():
    f = get_data_stooq("CLX26.F", start="20200101", end="20200115")
    assert f.shape[0] > 0


def test_get_data_stooq_dax():
    f = get_data_stooq("^DAX")
    assert f.shape[0] > 0


def test_stooq_googl():
    f = get_data_stooq("GOOGL.US")
    assert f.shape[0] > 0


def test_get_data_ibm():
    f = get_data_stooq("IBM.UK")
    assert f.shape[0] > 0



import os

import pytest


def pytest_addoption(parser):
    parser.addoption("--only-stable", action="store_true", help="run only stable tests")
    parser.addoption(
        "--skip-requires-api-key",
        action="store_true",
        help="skip tests that require an API key",
    )
    parser.addoption(
        "--strict-data-files",
        action="store_true",
        help="Fail if a test is skipped for missing data file.",
    )


def pytest_runtest_setup(item):
    if "stable" not in item.keywords and item.config.getoption("--only-stable"):
        pytest.skip("skipping due to --only-stable")

    if "requires_api_key" in item.keywords and item.config.getoption(
        "--skip-requires-api-key"
    ):
        pytest.skip("skipping due to --skip-requires-api-key")


@pytest.fixture
def datapath(request):
    """Get the path to a data file.

    Parameters
    ----------
    path : str
        Path to the file, relative to ``pandas/tests/``

    Returns
    -------
    path : path including ``pandas/tests``.

    Raises
    ------
    ValueError
        If the path doesn't exist and the --strict-data-files option is set.
    """
    BASE_PATH = os.path.join(os.path.dirname(__file__), "tests")

    def deco(*args):
        path = os.path.join(BASE_PATH, *args)
        if not os.path.exists(path):
            if request.config.getoption("--strict-data-files"):
                msg = "Could not find file {} and --strict-data-files is set."
                raise ValueError(msg.format(path))
            else:
                msg = "Could not find {}."
                pytest.skip(msg.format(path))
        return path

    return deco



from pandas_datareader.base import _BaseReader


class TSPReader(_BaseReader):

    """
    Returns DataFrame of historical TSP fund prices from symbols, over date
    range, start to end.

    Parameters
    ----------
    symbols : str, array-like object (list, tuple, Series), or DataFrame
        Single stock symbol (ticker), array-like object of symbols or
        DataFrame with index containing stock symbols.
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980'). Defaults to
        20 years before current date.
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    """

    all_symbols = frozenset(
        (
            "L Income",
            "L 2025",
            "L 2030",
            "L 2035",
            "L 2040",
            "L 2045",
            "L 2050",
            "L 2055",
            "L 2060",
            "L 2065",
            "G Fund",
            "F Fund",
            "C Fund",
            "S Fund",
            "I Fund",
        )
    )

    def __init__(
        self,
        symbols=all_symbols,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
    ):
        super(TSPReader, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )
        self._format = "string"

    @property
    def url(self):
        """API URL"""
        return "https://secure.tsp.gov/components/CORS/getSharePricesRaw.html"

    def read(self):
        """read one data from specified URL"""
        df = super(TSPReader, self).read()
        df.columns = map(lambda x: x.strip(), df.columns)
        df.drop(columns=self.all_symbols - set(self.symbols), inplace=True)
        return df

    @property
    def params(self):
        """Parameters to use in API calls"""
        return {
            "startdate": self.start.strftime("%Y%m%d"),
            "enddate": self.end.strftime("%Y%m%d"),
            "download": "0",
            "Lfunds": "1",
            "InvFunds": "1",
        }

    @staticmethod
    def _sanitize_response(response):
        """
        Clean up the response string
        """
        text = response.text.strip()
        if text[-1] == ",":
            return text[0:-1]
        return text



import datetime
import json
import os

from dateutil.relativedelta import relativedelta
import pandas as pd

from pandas_datareader.base import _DailyBaseReader

# Data provided for free by IEX
# Data is furnished in compliance with the guidelines promulgated in the IEX
# API terms of service and manual
# See https://iextrading.com/api-exhibit-a/ for additional information
# and conditions of use


class IEXDailyReader(_DailyBaseReader):

    """
    Returns DataFrame of historical stock prices
    from symbols, over date range, start to end. To avoid being penalized by
    IEX servers, pauses between downloading 'chunks' of symbols can be
    specified.

    Parameters
    ----------
    symbols : string, array-like object (list, tuple, Series), or DataFrame
        Single stock symbol (ticker), array-like object of symbols or
        DataFrame with index containing stock symbols.
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980'). Defaults to
        15 years before current date
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    chunksize : int, default 25
        Number of symbols to download consecutively before intiating pause.
    session : Session, default None
        requests.sessions.Session instance to be used
    api_key: str
        IEX Cloud Secret Token
    """

    def __init__(
        self,
        symbols=None,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
        chunksize=25,
        api_key=None,
    ):
        if api_key is None:
            api_key = os.getenv("IEX_API_KEY")
        if not api_key or not isinstance(api_key, str):
            raise ValueError(
                "The IEX Cloud API key must be provided either "
                "through the api_key variable or through the "
                " environment variable IEX_API_KEY"
            )
        # Support for sandbox environment (testing purposes)
        if os.getenv("IEX_SANDBOX") == "enable":
            self.sandbox = True
        else:
            self.sandbox = False
        self.api_key = api_key
        super(IEXDailyReader, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            chunksize=chunksize,
        )

    @property
    def default_start_date(self):
        today = datetime.date.today()
        return today - datetime.timedelta(days=365 * 15)

    @property
    def url(self):
        """API URL"""
        if self.sandbox is True:
            return "https://sandbox.iexapis.com/stable/stock/market/batch"
        else:
            return "https://cloud.iexapis.com/stable/stock/market/batch"

    @property
    def endpoint(self):
        """API endpoint"""
        return "chart"

    def _get_params(self, symbol):
        chart_range = self._range_string_from_date()
        if isinstance(symbol, list):
            symbolList = ",".join(symbol)
        else:
            symbolList = symbol
        params = {
            "symbols": symbolList,
            "types": self.endpoint,
            "range": chart_range,
            "token": self.api_key,
        }
        return params

    def _range_string_from_date(self):
        delta = relativedelta(self.start, datetime.datetime.now())
        years = delta.years * -1
        if 5 <= years <= 15:
            return "max"
        if 2 <= years < 5:
            return "5y"
        elif 1 <= years < 2:
            return "2y"
        elif 0 <= years < 1:
            delta_days = (datetime.datetime.now() - self.start).days
            if 0 <= delta_days < 6:
                return "5d"
            elif 6 <= delta_days < 28:
                return "1m"
            elif 28 <= delta_days < 84:
                return "3m"
            elif 84 <= delta_days < 168:
                return "6m"

            return "1y"
        else:
            raise ValueError("Invalid date specified. Must be within past 15 years.")

    def read(self):
        """Read data"""
        try:
            return self._read_one_data(self.url, self._get_params(self.symbols))
        finally:
            self.close()

    def _read_lines(self, out):
        data = out.read()
        json_data = json.loads(data)
        result = {}
        if type(self.symbols) is str:
            syms = [self.symbols]
        else:
            syms = self.symbols
        for symbol in syms:
            d = json_data.pop(symbol)["chart"]
            df = pd.DataFrame(d)
            df.set_index("date", inplace=True)
            values = ["open", "high", "low", "close", "volume"]
            df = df[values]
            sstart = self.start.strftime("%Y-%m-%d")
            send = self.end.strftime("%Y-%m-%d")
            df = df.loc[sstart:send]
            result.update({symbol: df})
        if len(result) > 1:
            result = pd.concat(result).unstack(level=0)
            result.columns.names = ["Attributes", "Symbols"]
            return result
        return result[self.symbols]



from datetime import datetime
from xml.etree import ElementTree

import numpy as np
from pandas import DataFrame, to_datetime
from six import string_types

from pandas_datareader.base import _DailyBaseReader


class NaverDailyReader(_DailyBaseReader):
    """Fetches daily historical data from Naver Finance.

    :param symbols: A single symbol; multiple symbols are not currently supported.
    :param adjust_price: Not implemented
    :param interval: Not implemented
    :param adjust_dividends: Not implemented
    """

    def __init__(
        self,
        symbols=None,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
        adjust_price=False,
        ret_index=False,
        chunksize=1,
        interval="d",
        get_actions=False,
        adjust_dividends=True,
    ):
        if not isinstance(symbols, string_types):
            raise NotImplementedError("Bulk-fetching is not implemented")

        super().__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            chunksize=chunksize,
        )

        self.headers = {
            "Sec-Fetch-Mode": "no-cors",
            "Referer": "https://finance.naver.com/item/fchart.nhn?code={}".format(
                symbols
            ),
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.100 Safari/537.36",  # noqa
        }

    @property
    def get_actions(self):
        return self._get_actions

    @property
    def url(self):
        return "https://fchart.stock.naver.com/sise.nhn"

    def _get_params(self, symbol):
        # NOTE: The server does not take start, end dates as inputs; it only
        # takes the number of trading days as an input. To circumvent this
        # pitfall, we calculate the number of business days between self.start
        # and the current date. And then we filter by self.end before returning
        # the final result (in _read_one_data()).
        days = np.busday_count(self.start.date(), datetime.now().date())
        params = {"symbol": symbol, "timeframe": "day", "count": days, "requestType": 0}
        return params

    def _read_one_data(self, url, params):
        """Read one data from specified symbol.

        :rtype: DataFrame
        """
        resp = self._get_response(url, params=params)
        parsed = self._parse_xml_response(resp.text)
        prices = DataFrame(
            parsed, columns=["Date", "Open", "High", "Low", "Close", "Volume"]
        )
        prices["Date"] = to_datetime(prices["Date"])
        prices = prices.set_index("Date")

        # NOTE: See _get_params() for explanations.
        return prices[(prices.index >= self.start) & (prices.index <= self.end)]

    def _parse_xml_response(self, xml_content):
        """Parses XML response from the server.

        An example of response:

            <?xml version="1.0" encoding="EUC-KR" ?>
            <protocol>
                <chartdata symbol="005930" name="Samsung Elctronics" count="500"
                        timeframe="day" precision="0" origintime="19900103">
                    <item data="20170918|218500|222000|217000|220500|72124" />
                    <item data="20170919|218000|221000|217500|219000|62753" />
                    ...
            </protocol>
        """
        root = ElementTree.fromstring(xml_content)
        items = root.findall("chartdata/item")

        for item in items:
            yield item.attrib["data"].split("|")



import pandas as pd

from pandas_datareader.base import _BaseReader
from pandas_datareader.compat import string_types
from pandas_datareader.io import read_jsdmx


class OECDReader(_BaseReader):
    """Get data for the given name from OECD."""

    _format = "json"

    @property
    def url(self):
        """API URL"""
        url = "http://stats.oecd.org/SDMX-JSON/data"

        if not isinstance(self.symbols, string_types):
            raise ValueError("data name must be string")

        # API: https://data.oecd.org/api/sdmx-json-documentation/
        return "{0}/{1}/all/all?".format(url, self.symbols)

    def _read_lines(self, out):
        """read one data from specified URL"""
        df = read_jsdmx(out)
        try:
            idx_name = df.index.name  # hack for pandas 0.16.2
            df.index = pd.to_datetime(df.index, errors="ignore")
            for col in df:
                df[col] = pd.to_numeric(df[col], errors="ignore")
            df = df.sort_index()
            df = df.truncate(self.start, self.end)
            df.index.name = idx_name
        except ValueError:
            pass
        return df



import os

import pandas as pd
import pytest

from pandas_datareader.tiingo import (
    TiingoDailyReader,
    TiingoIEXHistoricalReader,
    TiingoMetaDataReader,
    TiingoQuoteReader,
    get_tiingo_symbols,
)

pytestmark = pytest.mark.requires_api_key

TEST_API_KEY = os.getenv("TIINGO_API_KEY")
# Ensure blank TEST_API_KEY not used in pull request
TEST_API_KEY = None if not TEST_API_KEY else TEST_API_KEY

syms = ["GOOG", ["GOOG", "XOM"]]
ids = list(map(str, syms))


@pytest.fixture(params=syms, ids=ids)
def symbols(request):
    return request.param


@pytest.mark.skipif(TEST_API_KEY is None, reason="TIINGO_API_KEY not set")
def test_tiingo_quote(symbols):
    df = TiingoQuoteReader(symbols=symbols).read()
    assert isinstance(df, pd.DataFrame)
    if isinstance(symbols, str):
        symbols = [symbols]
    assert df.shape[0] == len(symbols)


@pytest.mark.skipif(TEST_API_KEY is None, reason="TIINGO_API_KEY not set")
def test_tiingo_historical(symbols):
    df = TiingoDailyReader(symbols=symbols).read()
    assert isinstance(df, pd.DataFrame)
    if isinstance(symbols, str):
        symbols = [symbols]
    assert df.index.levels[0].shape[0] == len(symbols)


@pytest.mark.skipif(TEST_API_KEY is None, reason="TIINGO_API_KEY not set")
def test_tiingo_iex_historical(symbols):
    df = TiingoIEXHistoricalReader(symbols=symbols).read()
    df.head()
    assert isinstance(df, pd.DataFrame)
    if isinstance(symbols, str):
        symbols = [symbols]
    assert df.index.levels[0].shape[0] == len(symbols)


@pytest.mark.skipif(TEST_API_KEY is None, reason="TIINGO_API_KEY not set")
def test_tiingo_metadata(symbols):
    df = TiingoMetaDataReader(symbols=symbols).read()
    assert isinstance(df, pd.DataFrame)
    if isinstance(symbols, str):
        symbols = [symbols]
    assert df.shape[1] == len(symbols)


def test_tiingo_no_api_key(symbols):
    from test.support import EnvironmentVarGuard

    env = EnvironmentVarGuard()
    env.unset("TIINGO_API_KEY")
    with env:
        with pytest.raises(ValueError):
            TiingoMetaDataReader(symbols=symbols)


@pytest.mark.skipif(
    pd.__version__ == "0.19.2", reason="pandas 0.19.2 does not like this file format"
)
def test_tiingo_stock_symbols():
    sym = get_tiingo_symbols()
    assert isinstance(sym, pd.DataFrame)



from datetime import datetime

import pytest

from pandas_datareader import DataReader


class TestNaver(object):
    @pytest.mark.parametrize(
        "symbol, start, end",
        [
            ("005930", (2019, 10, 1), (2019, 10, 7)),
            ("000660", (2018, 1, 1), (2018, 12, 31)),
            ("069500", (2017, 6, 3), (2018, 9, 9)),
        ],
    )
    def test_naver_daily_reader(self, symbol, start, end):
        start = datetime(*start)
        end = datetime(*end)
        reader = DataReader(symbol, "naver", start, end)

        assert reader.shape[1] == 5
        assert reader.index.min() >= start
        assert reader.index.max() <= end

    def test_bulk_fetch(self):
        with pytest.raises(NotImplementedError):
            DataReader(["005930", "000660"])



# pylint: disable-msg=E1101,W0613,W0603

import os

import numpy as np
import pandas as pd
from pandas import testing as tm
import pytest

from pandas_datareader.compat import PANDAS_0210
from pandas_datareader.io import read_jsdmx

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.mark.skipif(not PANDAS_0210, reason="Broken on old pandas")
def test_tourism(dirpath):
    # OECD -> Industry and Services -> Inbound Tourism
    result = read_jsdmx(os.path.join(dirpath, "jsdmx", "tourism.json"))
    assert isinstance(result, pd.DataFrame)
    jp = result["Japan"]
    visitors = [
        "China",
        "Hong Kong, China",
        "Total international arrivals",
        "Korea",
        "Chinese Taipei",
        "United States",
    ]

    exp_col = pd.Index(
        [
            "China",
            "Hong Kong, China",
            "Total international arrivals",
            "Korea",
            "Chinese Taipei",
            "United States",
        ],
        name="Variable",
    )
    exp_idx = pd.DatetimeIndex(
        [
            "2008-01-01",
            "2009-01-01",
            "2010-01-01",
            "2011-01-01",
            "2012-01-01",
            "2013-01-01",
            "2014-01-01",
            "2015-01-01",
            "2016-01-01",
        ],
        name="Year",
    )
    values = [
        [1000000.0, 550000.0, 8351000.0, 2382000.0, 1390000.0, 768000.0],
        [1006000.0, 450000.0, 6790000.0, 1587000.0, 1024000.0, 700000.0],
        [1413000.0, 509000.0, 8611000.0, 2440000.0, 1268000.0, 727000.0],
        [1043000.0, 365000.0, 6219000.0, 1658000.0, 994000.0, 566000.0],
        [1430000.0, 482000.0, 8368000.0, 2044000.0, 1467000.0, 717000.0],
        [1314000.0, 746000.0, 10364000.0, 2456000.0, 2211000.0, 799000.0],
        [2409000.0, 926000.0, 13413000.0, 2755000.0, 2830000.0, 892000.0],
        [4993689.0, 1524292.0, 19737409.0, 4002095.0, 3677075.0, 1033258.0],
        [6373564.0, 1839193.0, 24039700.0, 5090302.0, 4167512.0, 1242719.0],
    ]
    values = np.array(values, dtype="object")
    expected = pd.DataFrame(values, index=exp_idx, columns=exp_col)
    tm.assert_frame_equal(jp[visitors], expected)


@pytest.mark.skipif(not PANDAS_0210, reason="Broken on old pandas")
def test_land_use(dirpath):
    # OECD -> Environment -> Resources Land Use
    result = read_jsdmx(os.path.join(dirpath, "jsdmx", "land_use.json"))
    assert isinstance(result, pd.DataFrame)
    result = result.loc["2010":"2011"]

    cols = [
        "Arable land and permanent crops",
        "Arable and cropland % land area",
        "Total area",
        "Forest",
        "Forest % land area",
        "Land area",
        "Permanent meadows and pastures",
        "Meadows and pastures % land area",
        "Other areas",
        "Other % land area",
    ]
    exp_col = pd.MultiIndex.from_product(
        [["Japan", "United States"], cols], names=["Country", "Variable"]
    )
    exp_idx = pd.DatetimeIndex(["2010", "2011"], name="Year")
    values = [
        [
            53790.0,
            14.753154141525,
            377800.0,
            np.nan,
            np.nan,
            364600.0,
            5000.0,
            1.3713658804169,
            np.nan,
            np.nan,
            1897990.0,
            20.722767650476,
            9629090.0,
            np.nan,
            np.nan,
            9158960.0,
            2416000.0,
            26.378540795025,
            np.nan,
            np.nan,
        ],
        [
            53580.0,
            14.691527282698,
            377800.0,
            np.nan,
            np.nan,
            364700.0,
            5000.0,
            1.3709898546751,
            np.nan,
            np.nan,
            1897990.0,
            20.722767650476,
            9629090.0,
            np.nan,
            np.nan,
            9158960.0,
            2416000.0,
            26.378540795025,
            np.nan,
            np.nan,
        ],
    ]
    values = np.array(values)
    expected = pd.DataFrame(values, index=exp_idx, columns=exp_col)
    tm.assert_frame_equal(result[exp_col], expected)


@pytest.mark.skipif(not PANDAS_0210, reason="Broken on old pandas")
def test_quartervalue(dirpath):
    # https://stats.oecd.org/sdmx-json/data/QNA/AUS+AUT+BEL+CAN+CHL.GDP+B1_
    #    GE.CUR+VOBARSA.Q/all?startTime=2009-Q1&endTime=2011-Q4
    result = read_jsdmx(os.path.join(dirpath, "jsdmx", "oecd1.json"))
    assert isinstance(result, pd.DataFrame)
    expected = pd.DatetimeIndex(
        [
            "2009-01-01",
            "2009-04-01",
            "2009-07-01",
            "2009-10-01",
            "2010-01-01",
            "2010-04-01",
            "2010-07-01",
            "2010-10-01",
            "2011-01-01",
            "2011-04-01",
            "2011-07-01",
            "2011-10-01",
        ],
        dtype="datetime64[ns]",
        name=u"Period",
        freq=None,
    )
    tm.assert_index_equal(result.index, expected)



from distutils.version import LooseVersion
from functools import reduce
from io import StringIO
from urllib.error import HTTPError

import pandas as pd
from pandas.api.types import is_list_like, is_number
from pandas.io import common as com
from pandas.testing import assert_frame_equal

PANDAS_VERSION = LooseVersion(pd.__version__)

PANDAS_0210 = PANDAS_VERSION >= LooseVersion("0.21.0")
PANDAS_0220 = PANDAS_VERSION >= LooseVersion("0.22.0")
PANDAS_0230 = PANDAS_VERSION >= LooseVersion("0.23.0")

__all__ = [
    "HTTPError",
    "StringIO",
    "PANDAS_0210",
    "PANDAS_0220",
    "PANDAS_0230",
    "get_filepath_or_buffer",
    "str_to_bytes",
    "string_types",
    "assert_frame_equal",
    "is_list_like",
    "is_number",
    "lmap",
    "lrange",
    "concat",
    "reduce",
]


def get_filepath_or_buffer(filepath_or_buffer, encoding=None, compression=None):

    # Dictionaries are no longer considered valid inputs
    # for "get_filepath_or_buffer" starting in pandas >= 0.20.0
    if isinstance(filepath_or_buffer, dict):
        return filepath_or_buffer, encoding, compression
    try:
        tmp = com._get_filepath_or_buffer(
            filepath_or_buffer, encoding=encoding, compression=None
        )
        return tmp.filepath_or_buffer, tmp.encoding, tmp.compression
    except AttributeError:
        tmp = com.get_filepath_or_buffer(
            filepath_or_buffer, encoding=encoding, compression=None
        )
        return tmp


string_types = (str,)
binary_type = bytes


def str_to_bytes(s, encoding=None):
    if isinstance(s, bytes):
        return s
    return s.encode(encoding or "ascii")


def bytes_to_str(b, encoding=None):
    return b.decode(encoding or "utf-8")


def lmap(*args, **kwargs):
    return list(map(*args, **kwargs))


def lrange(*args, **kwargs):
    return list(range(*args, **kwargs))


def concat(*args, **kwargs):
    """
    Shim to wokr around sort keyword
    """
    if not PANDAS_0230 and "sort" in kwargs:
        del kwargs["sort"]
    return pd.concat(*args, **kwargs)



from datetime import datetime
import os

import pandas as pd
import pytest

from pandas_datareader import data as web
from pandas_datareader._utils import RemoteDataError

TEST_API_KEY = os.getenv("ALPHAVANTAGE_API_KEY")
TEST_API_KEY = None if not TEST_API_KEY else TEST_API_KEY

pytestmark = [
    pytest.mark.requires_api_key,
    pytest.mark.alpha_vantage,
    pytest.mark.skipif(TEST_API_KEY is None, reason="ALPHAVANTAGE_API_KEY not set"),
]


class TestAVTimeSeries(object):
    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")
        cls.col_index_adj = pd.Index(
            [
                "open",
                "high",
                "low",
                "close",
                "adjusted close",
                "volume",
                "dividend amount",
            ]
        )
        cls.col_index = pd.Index(["open", "high", "low", "close", "volume"])

    @property
    def start(self):
        return datetime(2015, 2, 9)

    @property
    def end(self):
        return datetime(2017, 5, 24)

    def test_av_bad_symbol(self):
        with pytest.raises((ValueError, RemoteDataError)):
            web.DataReader(
                "BADTICKER",
                "av-daily",
                start=self.start,
                end=self.end,
                retry_count=6,
                pause=20.5,
            )

    def test_av_daily(self):
        df = web.DataReader(
            "AAPL",
            "av-daily",
            start=self.start,
            end=self.end,
            retry_count=6,
            pause=20.5,
        )
        assert df.columns.equals(self.col_index)
        assert len(df) == 578
        assert df["volume"][-1] == 19118319

        expected1 = df.loc["2017-02-09"]
        assert expected1["close"] == 132.42
        assert expected1["high"] == 132.445

        expected2 = df.loc["2017-05-24"]
        assert expected2["close"] == 153.34
        assert expected2["high"] == 154.17

    def test_av_daily_adjusted(self):
        df = web.DataReader(
            "AAPL",
            "av-daily-adjusted",
            start=self.start,
            end=self.end,
            retry_count=6,
            pause=20.5,
        )
        assert df.columns.equals(
            pd.Index(
                [
                    "open",
                    "high",
                    "low",
                    "close",
                    "adjusted close",
                    "volume",
                    "dividend amount",
                    "split coefficient",
                ]
            )
        )
        assert len(df) == 578
        assert df["volume"][-1] == 19118319

        expected1 = df.loc["2017-02-09"]
        assert expected1["close"] == 132.42
        assert expected1["high"] == 132.445
        assert expected1["dividend amount"] == 0.57
        assert expected1["split coefficient"] == 1.0

        expected2 = df.loc["2017-05-24"]
        assert expected2["close"] == 153.34
        assert expected2["high"] == 154.17
        assert expected2["dividend amount"] == 0.00
        assert expected2["split coefficient"] == 1.0

    @staticmethod
    def _helper_df_weekly_monthly(df, adj=False):

        expected1 = df.loc["2015-02-27"]
        assert expected1["close"] == 128.46
        assert expected1["high"] == 133.60

        expected2 = df.loc["2017-03-31"]
        assert expected2["close"] == 143.66
        assert expected2["high"] == 144.5

    def test_av_weekly(self):
        df = web.DataReader(
            "AAPL",
            "av-weekly",
            start=self.start,
            end=self.end,
            retry_count=6,
            pause=20.5,
        )

        assert len(df) == 119
        assert df.iloc[0].name == "2015-02-13"
        assert df.iloc[-1].name == "2017-05-19"
        assert df.columns.equals(self.col_index)
        self._helper_df_weekly_monthly(df, adj=False)

    def test_av_weekly_adjusted(self):
        df = web.DataReader(
            "AAPL",
            "av-weekly-adjusted",
            start=self.start,
            end=self.end,
            retry_count=6,
            pause=20.5,
        )

        assert len(df) == 119
        assert df.iloc[0].name == "2015-02-13"
        assert df.iloc[-1].name == "2017-05-19"
        assert df.columns.equals(self.col_index_adj)
        self._helper_df_weekly_monthly(df, adj=True)

    def test_av_monthly(self):
        df = web.DataReader(
            "AAPL",
            "av-monthly",
            start=self.start,
            end=self.end,
            retry_count=6,
            pause=20.5,
        )

        assert len(df) == 27
        assert df.iloc[0].name == "2015-02-27"
        assert df.iloc[-1].name == "2017-04-28"
        assert df.columns.equals(self.col_index)
        self._helper_df_weekly_monthly(df, adj=False)

    def test_av_monthly_adjusted(self):
        df = web.DataReader(
            "AAPL",
            "av-monthly-adjusted",
            start=self.start,
            end=self.end,
            retry_count=6,
            pause=20.5,
        )

        assert df.columns.equals(self.col_index_adj)
        assert len(df) == 27
        assert df.iloc[0].name == "2015-02-27"
        assert df.iloc[-1].name == "2017-04-28"
        self._helper_df_weekly_monthly(df, adj=True)

    def test_av_intraday(self):
        # Not much available to test, but ensure close in length
        df = web.DataReader("AAPL", "av-intraday", retry_count=6, pause=20.5)

        assert len(df) >= 1
        assert "open" in df.columns
        assert "close" in df.columns

    def test_av_forex_daily(self):
        df = web.DataReader(
            "USD/JPY",
            "av-forex-daily",
            start=self.start,
            end=self.end,
            retry_count=6,
            pause=20.5,
        )
        assert df.columns.equals(self.col_index[:4])  # No volume col for forex
        assert len(df) == 598
        assert df.loc["2015-02-09"]["close"] == 118.6390
        assert df.loc["2017-05-24"]["high"] == 112.1290



import os
import sys

from ._version import get_versions
from .data import (
    DataReader,
    Options,
    get_components_yahoo,
    get_dailysummary_iex,
    get_data_alphavantage,
    get_data_enigma,
    get_data_famafrench,
    get_data_fred,
    get_data_moex,
    get_data_quandl,
    get_data_stooq,
    get_data_tiingo,
    get_data_yahoo,
    get_data_yahoo_actions,
    get_iex_book,
    get_iex_data_tiingo,
    get_iex_symbols,
    get_last_iex,
    get_markets_iex,
    get_nasdaq_symbols,
    get_quote_yahoo,
    get_recent_iex,
    get_records_iex,
    get_summary_iex,
    get_tops_iex,
)

PKG = os.path.dirname(__file__)

__version__ = get_versions()["version"]
del get_versions

__all__ = [
    "__version__",
    "get_components_yahoo",
    "get_data_enigma",
    "get_data_famafrench",
    "get_data_yahoo",
    "get_data_yahoo_actions",
    "get_quote_yahoo",
    "get_iex_book",
    "get_iex_symbols",
    "get_last_iex",
    "get_markets_iex",
    "get_recent_iex",
    "get_records_iex",
    "get_summary_iex",
    "get_tops_iex",
    "get_nasdaq_symbols",
    "get_data_quandl",
    "get_data_moex",
    "get_data_fred",
    "get_dailysummary_iex",
    "get_data_stooq",
    "DataReader",
    "Options",
    "get_data_tiingo",
    "get_iex_data_tiingo",
    "get_data_alphavantage",
    "test",
]


def test(extra_args=None):
    """
    Run the test suite

    Parameters
    ----------
    extra_args : {str, List[str]}
        A string or list of strings to pass to pytest. Default is
        ["--only-stable", "--skip-requires-api-key"]
    """
    try:
        import pytest
    except ImportError as err:
        raise ImportError("Need pytest>=5.0.1 to run tests") from err
    cmd = ["--only-stable", "--skip-requires-api-key"]
    if extra_args:
        if not isinstance(extra_args, list):
            extra_args = [extra_args]
        cmd = extra_args
    cmd += [PKG]
    joined = " ".join(cmd)
    print(f"running: pytest {joined}")
    sys.exit(pytest.main(cmd))



import numpy as np
import pandas as pd
from pandas import testing as tm
import pytest

from pandas_datareader import data as web
from pandas_datareader.famafrench import get_available_datasets

pytestmark = pytest.mark.stable


class TestFamaFrench(object):
    def test_get_data(self):
        keys = [
            "F-F_Research_Data_Factors",
            "F-F_ST_Reversal_Factor",
            "6_Portfolios_2x3",
            "Portfolios_Formed_on_ME",
            "Prior_2-12_Breakpoints",
            "ME_Breakpoints",
        ]

        for name in keys:
            ff = web.DataReader(name, "famafrench")
            assert "DESCR" in ff
            assert len(ff) > 1

    def test_get_available_datasets(self):
        pytest.importorskip("lxml")
        avail = get_available_datasets()
        assert len(avail) > 100

    def test_index(self):
        ff = web.DataReader("F-F_Research_Data_Factors", "famafrench")
        assert ff[0].index.freq == "M"
        assert ff[1].index.freq == "A-DEC"

    def test_f_f_research(self):
        results = web.DataReader(
            "F-F_Research_Data_Factors",
            "famafrench",
            start="2010-01-01",
            end="2010-12-01",
        )
        assert isinstance(results, dict)
        assert len(results) == 3

        exp = pd.DataFrame(
            {
                "Mkt-RF": [
                    -3.36,
                    3.40,
                    6.31,
                    2.00,
                    -7.89,
                    -5.57,
                    6.93,
                    -4.77,
                    9.54,
                    3.88,
                    0.60,
                    6.82,
                ],
                "SMB": [
                    0.37,
                    1.19,
                    1.44,
                    4.86,
                    0.14,
                    -1.84,
                    0.18,
                    -3.02,
                    3.93,
                    1.07,
                    3.78,
                    0.68,
                ],
                "HML": [
                    0.33,
                    3.19,
                    2.11,
                    2.91,
                    -2.39,
                    -4.52,
                    -0.36,
                    -1.90,
                    -3.23,
                    -2.46,
                    -0.95,
                    3.64,
                ],
                "RF": [
                    0.00,
                    0.00,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                ],
            },
            index=pd.period_range("2010-01-01", "2010-12-01", freq="M", name="Date"),
            columns=["Mkt-RF", "SMB", "HML", "RF"],
        )
        received = results[0]
        np.testing.assert_allclose(received, exp)
        tm.assert_index_equal(received.index, exp.index)
        tm.assert_index_equal(received.columns, exp.columns)

    def test_me_breakpoints(self):
        results = web.DataReader(
            "ME_Breakpoints", "famafrench", start="2010-01-01", end="2010-12-01"
        )
        assert isinstance(results, dict)
        assert len(results) == 2
        assert results[0].shape == (12, 21)

        exp_columns = pd.Index(
            [
                "Count",
                (0, 5),
                (5, 10),
                (10, 15),
                (15, 20),
                (20, 25),
                (25, 30),
                (30, 35),
                (35, 40),
                (40, 45),
                (45, 50),
                (50, 55),
                (55, 60),
                (60, 65),
                (65, 70),
                (70, 75),
                (75, 80),
                (80, 85),
                (85, 90),
                (90, 95),
                (95, 100),
            ],
            dtype="object",
        )
        tm.assert_index_equal(results[0].columns, exp_columns)

        exp_index = pd.period_range("2010-01-01", "2010-12-01", freq="M", name="Date")
        tm.assert_index_equal(results[0].index, exp_index)

    def test_prior_2_12_breakpoints(self):
        results = web.DataReader(
            "Prior_2-12_Breakpoints", "famafrench", start="2010-01-01", end="2010-12-01"
        )
        assert isinstance(results, dict)
        assert len(results) == 2
        assert results[0].shape == (12, 22)

        exp_columns = pd.Index(
            [
                "<=0",
                ">0",
                (0, 5),
                (5, 10),
                (10, 15),
                (15, 20),
                (20, 25),
                (25, 30),
                (30, 35),
                (35, 40),
                (40, 45),
                (45, 50),
                (50, 55),
                (55, 60),
                (60, 65),
                (65, 70),
                (70, 75),
                (75, 80),
                (80, 85),
                (85, 90),
                (90, 95),
                (95, 100),
            ],
            dtype="object",
        )
        tm.assert_index_equal(results[0].columns, exp_columns)

        exp_index = pd.period_range("2010-01-01", "2010-12-01", freq="M", name="Date")
        tm.assert_index_equal(results[0].index, exp_index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

import versioneer

NAME = "pandas-datareader"


def readme():
    with open("README.md") as f:
        return f.read()


install_requires = []
with open("./requirements.txt") as f:
    install_requires = f.read().splitlines()
with open("./requirements-dev.txt") as f:
    tests_require = f.read().splitlines()

setup(
    name=NAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Data readers extracted from the pandas codebase,"
    "should be compatible with recent pandas versions",
    long_description=readme(),
    license="BSD License",
    author="The PyData Development Team",
    author_email="pydata@googlegroups.com",
    url="https://github.com/pydata/pandas-datareader",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
    keywords="data",
    install_requires=install_requires,
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    test_suite="tests",
    tests_require=tests_require,
    zip_safe=False,
    python_requires=">=3.6",
)



import os

import pandas as pd

from pandas_datareader._utils import RemoteDataError
from pandas_datareader.base import _BaseReader

AV_BASE_URL = "https://www.alphavantage.co/query"


class AlphaVantage(_BaseReader):
    """
    Base class for all Alpha Vantage queries

    Notes
    -----
    See `Alpha Vantage <https://www.alphavantage.co/>`__
    """

    _format = "json"

    def __init__(
        self,
        symbols=None,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
        api_key=None,
    ):
        super(AlphaVantage, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )
        if api_key is None:
            api_key = os.getenv("ALPHAVANTAGE_API_KEY")
        if not api_key or not isinstance(api_key, str):
            raise ValueError(
                "The AlphaVantage API key must be provided "
                "either through the api_key variable or "
                "through the environment variable "
                "ALPHAVANTAGE_API_KEY"
            )
        self.api_key = api_key

    @property
    def url(self):
        """API URL"""
        return AV_BASE_URL

    @property
    def params(self):
        return {"function": self.function, "apikey": self.api_key}

    @property
    def function(self):
        """Alpha Vantage endpoint function"""
        raise NotImplementedError

    @property
    def data_key(self):
        """Key of data returned from Alpha Vantage"""
        raise NotImplementedError

    def _read_lines(self, out):
        try:
            df = pd.DataFrame.from_dict(out[self.data_key], orient="index")
        except KeyError:
            if "Error Message" in out:
                raise ValueError(
                    "The requested symbol {} could not be "
                    "retrieved. Check valid ticker"
                    ".".format(self.symbols)
                )
            else:
                raise RemoteDataError(
                    " Their was an issue from the data vendor "
                    "side, here is their response: {}".format(out)
                )
        df = df[sorted(df.columns)]
        df.columns = [id[3:] for id in df.columns]
        return df



from datetime import datetime, timedelta

import pandas as pd

from pandas_datareader.exceptions import UnstableAPIWarning
from pandas_datareader.iex import IEX

# Data provided for free by IEX
# Data is furnished in compliance with the guidelines promulgated in the IEX
# API terms of service and manual
# See https://iextrading.com/api-exhibit-a/ for additional information
# and conditions of use


class DailySummaryReader(IEX):
    """
    Daily statistics from IEX for a day or month
    """

    def __init__(
        self, symbols=None, start=None, end=None, retry_count=3, pause=0.1, session=None
    ):
        import warnings

        warnings.warn(
            "Daily statistics is not working due to issues with the " "IEX API",
            UnstableAPIWarning,
        )
        self.curr_date = start
        super(DailySummaryReader, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )

    @property
    def service(self):
        """Service endpoint"""
        return "stats/historical/daily"

    def _get_params(self, symbols):
        p = {}

        if self.curr_date is not None:
            p["date"] = self.curr_date.strftime("%Y%m%d")

        return p

    def read(self):
        """Unfortunately, IEX's API can only retrieve data one day or one month
        at a time. Rather than specifying a date range, we will have to run
        the read function for each date provided.

        :return: DataFrame
        """
        tlen = self.end - self.start
        dfs = []
        for date in (self.start + timedelta(n) for n in range(tlen.days)):
            self.curr_date = date
            tdf = super(DailySummaryReader, self).read()
            dfs.append(tdf)
        return pd.concat(dfs)


class MonthlySummaryReader(IEX):
    """Monthly statistics from IEX"""

    def __init__(
        self, symbols=None, start=None, end=None, retry_count=3, pause=0.1, session=None
    ):
        self.curr_date = start
        self.date_format = "%Y%m"

        super(MonthlySummaryReader, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )

    @property
    def service(self):
        """Service endpoint"""
        return "stats/historical"

    def _get_params(self, symbols):
        p = {}

        if self.curr_date is not None:
            p["date"] = self.curr_date.strftime(self.date_format)

        return p

    def read(self):
        """Unfortunately, IEX's API can only retrieve data one day or one month
         at a time. Rather than specifying a date range, we will have to run
         the read function for each date provided.

        :return: DataFrame
        """
        tlen = self.end - self.start
        dfs = []

        # Build list of all dates within the given range
        lrange = [x for x in (self.start + timedelta(n) for n in range(tlen.days))]

        mrange = []
        for dt in lrange:
            if datetime(dt.year, dt.month, 1) not in mrange:
                mrange.append(datetime(dt.year, dt.month, 1))
        lrange = mrange

        for date in lrange:
            self.curr_date = date
            tdf = super(MonthlySummaryReader, self).read()

            # We may not return data if this was a weekend/holiday:
            if not tdf.empty:
                tdf["date"] = date.strftime(self.date_format)
                dfs.append(tdf)

        # We may not return any data if we failed to specify useful parameters:
        return pd.concat(dfs) if len(dfs) > 0 else pd.DataFrame()


class RecordsReader(IEX):
    """
    Total matched volume information from IEX
    """

    def __init__(
        self, symbols=None, start=None, end=None, retry_count=3, pause=0.1, session=None
    ):
        super(RecordsReader, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )

    @property
    def service(self):
        """Service endpoint"""
        return "stats/records"

    def _get_params(self, symbols):
        # Record Stats API does not take any parameters, returning empty dict
        return {}


class RecentReader(IEX):
    """Recent trading volume from IEX

    Notes
    -----
    Returns 6 fields for each day:

      * date: refers to the trading day.
      * volume: refers to executions received from order routed to away
        trading centers.
      * routedVolume: refers to single counted shares matched from executions
        on IEX.
      * marketShare: refers to IEX's percentage of total US Equity market
        volume.
      * isHalfday: will be true if the trading day is a half day.
      * litVolume: refers to the number of lit shares traded on IEX
        (single-counted).
    """

    def __init__(
        self, symbols=None, start=None, end=None, retry_count=3, pause=0.1, session=None
    ):
        super(RecentReader, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )

    @property
    def service(self):
        """Service endpoint"""
        return "stats/recent"

    def _get_params(self, symbols):
        # Record Stats API does not take any parameters, returning empty dict
        return {}



import pytest
from requests.exceptions import HTTPError

import pandas_datareader as pdr
from pandas_datareader import data as web
from pandas_datareader.exceptions import ImmediateDeprecationError

pytestmark = pytest.mark.requires_api_key

TEST_API_KEY = "DEPRECATED"


class TestEnigma(object):
    @property
    def dataset_id(self):
        """
        USDA Food Recall Archive - 1996
        Selected for being a relatively small dataset.
        https://public.enigma.com/datasets/292129b0-1275-44c8-a6a3-2a0881f24fe1
        """
        return "292129b0-1275-44c8-a6a3-2a0881f24fe1"

    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")

    def test_enigma_datareader(self):
        try:
            with pytest.raises(ImmediateDeprecationError):
                web.DataReader(self.dataset_id, "enigma", api_key=TEST_API_KEY)
        except HTTPError as e:
            pytest.skip(e)

    def test_enigma_get_data_enigma(self):
        try:
            with pytest.raises(ImmediateDeprecationError):
                pdr.get_data_enigma(self.dataset_id, TEST_API_KEY)
        except HTTPError as e:
            pytest.skip(e)

    def test_bad_key(self):
        with pytest.raises(ImmediateDeprecationError):
            web.DataReader(self.dataset_id, "enigma", api_key=TEST_API_KEY + "xxx")

    def test_bad_dataset_id(self):
        with pytest.raises(ImmediateDeprecationError):
            web.DataReader(
                "zzzzzzzz-zzzz-zzzz-zzzz-zzzzzzzzzzz", "enigma", api_key=TEST_API_KEY
            )



import datetime as dt

import pandas as pd

from pandas_datareader.base import _DailyBaseReader
from pandas_datareader.compat import StringIO, binary_type, concat, is_list_like


class MoexReader(_DailyBaseReader):
    """
    Returns a DataFrame of historical stock prices from symbols from Moex

    Parameters
    ----------
    symbols : str, an array-like object (list, tuple, Series), or a DataFrame
        A single stock symbol (secid), an array-like object of symbols or
        a DataFrame with an index containing stock symbols.
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980'). Defaults to
        20 years before current date.
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        The number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    chunksize : int, default 25
        The number of symbols to download consecutively before intiating pause.
    session : Session, default None
        requests.sessions.Session instance to be used

    Notes
    -----
    To avoid being penalized by Moex servers, pauses more than 0.1s between
    downloading 'chunks' of symbols can be specified.
    """

    def __init__(self, *args, **kwargs):
        super(MoexReader, self).__init__(*args, **kwargs)
        self.start = self.start.date()
        self.end_dt = self.end
        self.end = self.end.date()

        if isinstance(self.symbols, pd.DataFrame):
            self.symbols = self.symbols.index.tolist()
        elif not is_list_like(self.symbols):
            self.symbols = [self.symbols]

        self.__markets_n_engines = {}  # dicts for tuples of engines and markets

    __url_metadata = "https://iss.moex.com/iss/securities/{symbol}.csv"
    __url_data = (
        "https://iss.moex.com/iss/history/engines/{engine}/"
        "markets/{market}/securities/{symbol}.csv"
    )

    @property
    def url(self):
        """Return a list of API URLs per symbol"""

        if not self.__markets_n_engines:
            raise Exception(
                "Accessing url property before invocation "
                "of read() or _get_metadata() methods"
            )

        return [
            self.__url_data.format(engine=engine, market=market, symbol=s)
            for s in self.symbols
            if s in self.__markets_n_engines
            for market, engine in self.__markets_n_engines[s]
        ]

    def _get_params(self, start):
        """Return a dict for REST API GET request parameters"""

        params = {
            "iss.only": "history",
            "iss.dp": "point",
            "iss.df": "%Y-%m-%d",
            "iss.tf": "%H:%M:%S",
            "iss.dtf": "%Y-%m-%d %H:%M:%S",
            "iss.json": "extended",
            "callback": "JSON_CALLBACK",
            "from": start,
            "till": self.end_dt.strftime("%Y-%m-%d"),
            "limit": 100,
            "start": 0,
            "sort_order": "TRADEDATE",
            "sort_order_desc": "asc",
        }
        return params

    def _get_metadata(self):
        """Get markets and engines for the given symbols"""

        markets_n_engines = {}
        boards = {}

        for symbol in self.symbols:
            response = self._get_response(self.__url_metadata.format(symbol=symbol))
            text = self._sanitize_response(response)
            if len(text) == 0:
                service = self.__class__.__name__
                raise IOError(
                    "{} request returned no data; check URL for invalid "
                    "inputs: {}".format(service, self.__url_metadata)
                )
            if isinstance(text, binary_type):
                text = text.decode("windows-1251")

            header_str = "secid;boardid;"
            get_data = False
            for s in text.splitlines():
                if s.startswith(header_str):
                    get_data = True
                    continue
                if get_data and s != "":
                    fields = s.split(";")

                    if symbol not in markets_n_engines:
                        markets_n_engines[symbol] = list()

                    markets_n_engines[symbol].append(
                        (fields[5], fields[7])
                    )  # market and engine

                    if fields[14] == "1":  # main board for symbol
                        symbol_U = symbol.upper()
                        boards[symbol_U] = fields[1]

            if symbol not in markets_n_engines:
                raise IOError(
                    "{} request returned no metadata: {}\n"
                    "Typo in the security symbol `{}`?".format(
                        self.__class__.__name__,
                        self.__url_metadata.format(symbol=symbol),
                        symbol,
                    )
                )
            if symbol in markets_n_engines:
                markets_n_engines[symbol] = list(set(markets_n_engines[symbol]))
        return markets_n_engines, boards

    def read_all_boards(self):
        """Read all data from every board for every ticker"""

        markets_n_engines, boards = self._get_metadata()
        try:
            self.__markets_n_engines = markets_n_engines

            urls = self.url  # generate urls per symbols
            dfs = []  # an array of pandas dataframes per symbol to concatenate

            for i in range(len(urls)):
                out_list = []
                date_column = None

                while True:  # read in a loop with small date intervals
                    if len(out_list) > 0:
                        if date_column is None:
                            date_column = out_list[0].split(";").index("TRADEDATE")

                        # get the last downloaded date
                        start_str = out_list[-1].split(";", 4)[date_column]
                        start = dt.datetime.strptime(start_str, "%Y-%m-%d").date()
                    else:
                        start_str = self.start.strftime("%Y-%m-%d")
                        start = self.start

                    if start > self.end or start > dt.date.today():
                        break

                    params = self._get_params(start_str)
                    strings_out = self._read_url_as_String(
                        urls[i], params
                    ).splitlines()[2:]
                    strings_out = list(filter(lambda x: x.strip(), strings_out))

                    if len(out_list) == 0:
                        out_list = strings_out
                        if len(strings_out) < 101:  # all data received - break
                            break
                    else:
                        out_list += strings_out[1:]  # remove a CSV head line
                        if len(strings_out) < 100:  # all data received - break
                            break

                if len(out_list) > 0:
                    str_io = StringIO("\r\n".join(out_list))
                    dfs.append(self._read_lines(str_io))  # add a new DataFrame
        finally:
            self.close()

        if len(dfs) == 0:
            raise IOError(
                "{} returned no data; "
                "check URL or correct a date".format(self.__class__.__name__)
            )
        elif len(dfs) > 1:
            b = concat(dfs, axis=0, join="outer", sort=True)
        else:
            b = dfs[0]
        return b

    def read(self):
        """Read data from the primary board for each ticker"""
        markets_n_engines, boards = self._get_metadata()
        b = self.read_all_boards()
        result = pd.DataFrame()
        for secid in list(set(b["SECID"].tolist())):
            part = b[b["BOARDID"] == boards[secid]]
            result = result.append(part)
        result = result.drop_duplicates()
        return result

    def _read_url_as_String(self, url, params=None):
        """Open an url (and retry)"""

        response = self._get_response(url, params=params)
        text = self._sanitize_response(response)
        if len(text) == 0:
            service = self.__class__.__name__
            raise IOError(
                "{} request returned no data; check URL for invalid "
                "inputs: {}".format(service, self.url)
            )
        if isinstance(text, binary_type):
            text = text.decode("windows-1251")
        return text

    def _read_lines(self, input):
        """Return a pandas DataFrame from input"""

        return pd.read_csv(
            input,
            index_col="TRADEDATE",
            parse_dates=True,
            sep=";",
            na_values=("-", "null"),
        )
